'   sdspi:  SPI interface to a Secure Digital card.
'
'   Copyright 2008    Radical Eye Software
'   Modified  2009 by Michael Park: added pointer to rendezvous variables.
'   Modified  2010 by Michael Green: modified for use with C3 Board
'   Modified  2010 by Michael Green: added SPI sendRecv operation
'   Modified  2010 by Michael Green: added boot operation for flash
'
'   See end of file for terms of use.
'
'   You probably never want to call this; you want to use sxfs
'   instead (which calls this); this is only the lowest layer.
'
'   Most of this code is not general-purpose SPI code; it's very
'   specific to reading SD cards, although it can be used as an
'   example.
'
'   The code does not use CRC at the moment (this is the default).
'   With some additional effort we can probe the card to see if it
'   supports CRC, and if so, turn it on.   
'
'   All operations are guarded by a watchdog timer, just in case
'   no card is plugged in or something else is wrong.  If an
'   operation does not complete in one second it is aborted.
'
'   sendRecv is general purpose SPI code for use with SPI SRAM,
'   SPI Flash, and SPI ADCs.  It sends a 1 to 4 byte command and
'   its argument(s) then sends or receives an optional block of
'   data.
'
con
   sectorshift = 9                     ' microSD card sector size power of 2
   sectorsize  = |< sectorshift        ' microSD card sector size in bytes

   FLASH       = 3                     ' Atmel or Winbond Flash /CS #
   RdData      = $03_000000            ' Read command: Data / Addr / (Data)

   clkfreqVal  = $0000                 ' Current CLKFREQ value stored here
   clksetVal   = $0004                 ' Current CLKSET value stored here
   vbase       = $0008                 ' Length of Spin program loaded (# longs * 4)
   dbase       = $000A                 ' Address of start of stack (marker below)
   
var
   long cog
   long pCommand, pParam, pBlockno     ' rendezvous between spin and assembly

pub stop
   if cog
      cogstop(cog~ - 1)

pub start(basepin, pRendezvous) | i
'
'   Initialize the card!  Send a whole bunch of
'   clocks (in case the previous program crashed
'   in the middle of a read command or something),
'   then a reset command, and then wait until the
'   card goes idle.
'
'   If basepin is zero, the default (C3) pin assignments are used.
'   If basepin is non-zero, it contains the I/O pin numbers, each
'   in 5 bits.  From LSB these are di, do, clk, csClr, and csCnt.
'   Bits 27-25 contain the cs select code for the SD card.
'
   if basepin                      ' change I/O pins if requested
      repeat i from 0 to 4         ' order: di, do, Clk, csClr, csCnt
         diMask[i] := |< (di[i] := basepin & $1F)
         basepin >>= 5             ' pin number and bit mask for each
      selNo := basepin & $07       ' last item (selNo) is 3 bits
   stop

   clockfreq := clkfreq            ' adjust timeout based on clock   
   pCommand := pRendezvous
   pParam := pRendezvous + 4
   pBlockno := pRendezvous + 8
   
   long[pCommand] := "I"
   cog := 1 + cognew(@entry, pRendezvous)
   repeat while long[pCommand]
   if long[pParam]
      abort long[pParam]
   return 0
   
pub readblock(n, b)
'
'   Read a single block.  The "n" passed in is the
'   block number (blocks are 512 bytes); the b passed
'   in is the address of 512 blocks to fill with the
'   data.
'                              
   long[pParam] := b               ' save buffer address
   long[pBlockno] := n             ' save block number
   long[pCommand] := "R"           ' initiate read
   repeat while long[pCommand]     ' wait for read to finish
   if long[pParam]
      abort long[pParam]           ' abort on error
   return 0
   
pub writeblock(n, b)
'
'   Write a single block.  Mirrors the read above.
'
   long[pParam] := b               ' save buffer address
   long[pBlockno] := n             ' save block number
   long[pCommand] := "W"           ' initiate write
   repeat while long[pCommand]     ' wait for write to finish
   if long[pParam]
      abort long[pParam]           ' abort on error
   return 0
   
pub sendRecv(d, sD, rCt, rA) | i, m    ' Send and possibly receive
'  d   = Device to be used: 1 - SRAM0, 2 - SRAM1, 3 - Flash Memory
'                           4 - ADC,   6 - CS 6,  7 - CS 7
'  sD  = Value to be sent MSB first, right justified in parameter
'  rCt = Number of bytes to be transferred (>0 received, <0 sent)
'  rA  = 0 or address of buffer area.  If 0, @RESULT is used
   if rA                           ' Byte count in most significant word
      long[pParam] := (rCt << 16) + rA
   else                            ' Address in least significant word
      long[pParam] := (rCt << 16) + @result
   long[pBlockno] := sD            ' Possible literal value
   long[pCommand] := "0" + d       ' Initiate operation on specified device
   repeat while long[pCommand]     ' Wait until done (no error checking)

dat
                        org     0
entry                   mov     comptr,par              ' point to command
                        mov     parptr,comptr           ' pointer to first parameter
                        add     parptr,#4
                        mov     parptr2,parptr          ' pointer to second parameter
                        add     parptr2,#4
                        mov     outa,#0
                        andn    dira,doMask             ' input is: do
                        or      dira,diMask             ' outputs are: di, clk, clr, cnt
                        or      dira,clkMask            ' note: select counter cleared
                        or      dira,clrMask
                        or      dira,cntMask
                        mov     phsa,#0
                        mov     frqa,#0
                        mov     acca,nco                ' initialize counter A
                        add     acca,clk                '  clock low, don't count
                        mov     ctra,acca
                        neg     phsb,#1
                        mov     frqb,#0
                        mov     acca,nco                ' initialize counter B
                        add     acca,di                 '  data high, never count
                        mov     ctrb,acca
                        mov     ctr2,onek
oneloop                 call    #sendiohi               ' send 1K*8 clocks
                        djnz    ctr2,#oneloop
                        mov     starttime,cnt
                        mov     cmdo,#0                 ' reset command
                        mov     cmdp,#0
                        call    #cmdSel
                        andn    outa,clrMask            ' deselect SPI device
                        call    #sendiohi
initloop                mov     cmdo,#55                ' app_cmd
                        call    #cmdSel
                        mov     cmdo,#41                ' send_op_cond
                        call    #cmd
                        andn    outa,clrMask            ' deselect SPI device
                        cmp     accb,#1 wz
                 if_z   jmp     #initloop
                        wrlong  accb,parptr
finished                mov     frqa,#0                 ' mark operation done
                        wrlong  frqa,comptr             '  and stop SPI clock
                        andn    outa,clrMask            ' deselect SPI device
                        neg     phsb,#1                 ' send out 8 clocks to
                        call    #sendiohi               '  make sure bus off
                        mov     frqa,#0                 ' turn off SPI clock
                        mov     phsa,#0                 '  and leave clock low
pause                   mov     acca,#511
                        add     acca,cnt                ' wait about 2ms
                        waitcnt acca,#0
waitloop                rdlong  acca,comptr wz
                        cmp     acca,#"B" wz            ' byte i/o
                 if_z   jmp     #byteio
                        cmp     acca,#"R" wz            ' read a block of data
                 if_z   jmp     #rblock
                        cmp     acca,#"W" wz            ' write a block of data
                 if_z   jmp     #wblock
                        cmp     acca,#"a" wc            ' initiate boot operation
                 if_c   jmp     #:skipTest              '  "a"-"h" uses a cog 0-7
                        cmp     acca,#"i" wc,wz         '  "i" stops all cogs first
                 if_c   jmp     #loadFile
                 if_z   jmp     #bootFile
:skipTest               sub     acca,#"0" wc            ' read / write a block
                 if_c   jmp     #pause                  '  of data to other SPI
                        cmp     acca,#8 wc              '   devices
                if_nc   jmp     #pause
spiSendRecv             andn    outa,clrMask            ' make sure SPI deselected
                        or      outa,clrMask
                        rdlong  ctr2,parptr             ' fetch byte count & address
                        mov     ctr,ctr2
                        sar     ctr2,#16                ' extract byte count
                        and     ctr,wordMask            '  and address
:cntCS                  or      outa,cntMask            ' pulse csCnt acca times
                        andn    outa,cntMask
                        djnz    acca,#:cntCS
                        rdlong  acca,parptr2            ' fetch SPI command & data
                        tjz     acca,#skipImm           ' skip if all zeros
                        mov     selCnt,#4               ' send 1-4 bytes
                        test    wordMask,#1 wc          ' shift in one bits
:scanMask               test    acca,msbMask wz         ' left justify the data
                 if_z   rcl     acca,#8                 '  and keep byte count
                 if_z   djnz    selCnt,#:scanMask       ' count won't go zero
                        mov     phsb,acca
:sendImm                call    #sendRecvBits           ' send MSB and fill with ones
                        djnz    selCnt,#:sendImm
skipImm                 test    ctr2,signMask wc        ' If byte count < 0
                if_nc   jmp     #skipSend               '  then transmit data
                        neg     ctr2,ctr2               ' Transmit count positive
:sendData               test    wordMask,#1 wc          ' Force carry true
                        rdbyte  phsb,ctr                ' Get the data byte
                        rcl     phsb,#24                ' Position and fill with ones
                        add     ctr,#1                  ' Increment address
                        call    #sendRecvBits           ' Send the data
                        djnz    ctr2,#:sendData         ' Continue to next byte
skipSend                tjz     ctr2,#skipRecv          ' Skip if count == 0
:recvByte               neg     phsb,#1                 ' Send all one bits
                        call    #sendRecvBits           ' Receive data
                        wrbyte  phsb,ctr                ' Store the data byte
                        add     ctr,#1                  ' Increment address
                        djnz    ctr2,#:recvByte         ' Continue to next byte
skipRecv                jmp     #finished

sendRecvBits            mov     bits,#8                 ' MSB of data on output pin
:sendRecvBit            test    domask,ina wc           ' Read input bit
                        neg     phsa,#1                 ' SPI clock high for 100ns
                        rcl     phsb,#1                 ' Shift in a one bit
                        mov     phsa,#0                 ' SPI clock low for 150ns
                        djnz    bits,#:sendRecvBit
sendRecvBits_ret        ret

wblock                  mov     starttime,cnt
                        mov     ctr2,sector
                        mov     cmdo,#24
                        rdlong  cmdp,parptr2
                        call    #cmdSel
                        mov     phsb,#$fe
                        call    #sendio
                        rdlong  accb,parptr
                        neg     frqa,#1
wbyte                   rdbyte  phsb,accb
                        shl     phsb,#23
                        add     accb,#1
                        mov     ctr,#8
wbit                    mov     phsa,#8
                        shl     phsb,#1
                        djnz    ctr,#wbit
                        djnz    ctr2,#wbyte
                        neg     phsb,#1
                        call    #sendiohi
                        call    #sendiohi
                        call    #readresp
                        and     accb,#$1f
                        sub     accb,#5
                        wrlong  accb,parptr
                        call    #busy
                        jmp     #finished

rblock                  mov     starttime,cnt
                        mov     ctr2,sector
                        mov     cmdo,#17
                        rdlong  cmdp,parptr2
                        call    #cmdSel
                        call    #readresp
                        rdlong  accb,parptr
                        sub     accb,#1
rbyte                   mov     phsa,hifreq
                        mov     frqa,freq
                        add     accb,#1
                        test    domask,ina wc
                        addx    acca,acca
                        test    domask,ina wc
                        addx    acca,acca
                        test    domask,ina wc
                        addx    acca,acca
                        test    domask,ina wc
                        addx    acca,acca
                        test    domask,ina wc
                        addx    acca,acca
                        test    domask,ina wc
                        addx    acca,acca
                        test    domask,ina wc
                        addx    acca,acca
                        mov     frqa,#0
                        test    domask,ina wc
                        addx    acca,acca
                        wrbyte  acca,accb
                        djnz    ctr2,#rbyte        
                        mov     frqa,#0                 ' stop SPI clock
                        neg     phsb,#1                 ' set SPI MOSI to high
                        call    #sendiohi               ' now send 16 clocks
                        call    #sendiohi
                        andn    outa,clrMask            ' deselect SPI device
                        wrlong  ctr2,parptr
                        jmp     #finished

byteio                  rdlong  phsb,parptr             ' send byte parameter
                        call    #sendio
                        wrlong  accb,parptr             ' receive a byte too
                        jmp     #finished
 
sendio                  rol     phsb,#24                ' phsa would be initially zero
sendiohi                mov     bits,#8                 ' frqa set here to -1.  This makes
                        neg     frqa,#1                 '  SPI clock high for about 8 system
                        mov     accb,#0                 '   clocks.  Setting phsa to 8 then
bit                     mov     phsa,#8                 '    takes SPI clock low for 9 system
                        test    domask,ina wc           '     clocks, then high again for 20
                        addx    accb,accb
                        rol     phsb,#1
                        djnz    bits,#bit
sendiohi_ret
sendio_ret              ret

checktime               mov     duration,cnt
                        sub     duration,starttime
                        cmp     duration,clockfreq wc
checktime_ret    if_c   ret
                        neg     duration,#13
                        wrlong  duration,parptr
                        jmp     #finished

cmdSel                  andn    outa,clrMask             ' make sure SPI deselected
                        or      outa,clrMask
                        mov     selCnt,selNo
:cntCS                  or      outa,cntMask             ' pulse csCnt selNo times
                        andn    outa,cntMask
                        djnz    selCnt,#:cntCS
cmd                     neg     phsb,#1
                        call    #sendiohi                ' send 8 clocks, data high
                        mov     phsb,cmdo
                        add     phsb,#$40                ' send command
                        call    #sendio
                        mov     phsb,cmdp                ' get sector number and
                        shl     phsb,#9                  '  send as byte address
                        call    #sendiohi
                        call    #sendiohi
                        call    #sendiohi
                        call    #sendiohi
                        mov     phsb,#$95                ' send default CRC
                        call    #sendio
readresp                neg     phsb,#1
                        call    #sendiohi                ' read response to accb
                        call    #checktime
                        cmp     accb,#$ff wz
                 if_z   jmp     #readresp 
cmd_ret
cmdSel_ret
readresp_ret            ret

busy                    neg     phsb,#1
                        call    #sendiohi
                        call    #checktime
                        cmp     accb,#$0 wz
                 if_z   jmp     #busy
busy_ret                ret

bootFile                mov     ctr,#7                  ' go through all the cogs
                        cogid   startSpin               '  looking for current cog
:killCogs               cmp     startSpin,ctr wz
                 if_nz  cogstop ctr                     ' if match fails, stop cog
                        djnz    ctr,#:killCogs
                        cmp     startSpin,ctr wz
                 if_nz  cogstop ctr
                        neg     changeClk,#1            ' change system clock to
                        jmp     #startLoad              '  new program's values
loadFile                mov     startSpin,acca          ' caller gives cog to use
                        sub     startSpin,#"a"
                        mov     changeClk,#0            ' leave system clock alone
startLoad               or      startSpin,interpreter
                        rdlong  progSize,parptr         ' size of program (bytes)
                        rdlong  bootSect,parptr2        ' start of file
                        mov     bootAddr,#0             ' start at location zero
                        rdlong  oldClkFrq,#clkfreqVal   ' save CLKFREQ value
                        rdbyte  oldClkMod,#clksetVal    '  and CLKSET value
bootRead                mov     acca,flashSect          ' adjust length as needed
                        max     acca,progSize           '  for size of program
                        sub     progSize,acca
                        mov     phsb,bootSect           ' offset to data area
                        add     phsb,#16
                        call    #cardSetup              ' read from flash memory
:readByte               neg     phsb,#1
                        call    #sendRecvBits           ' read a byte and
                        wrbyte  phsb,bootAddr           '  store in hub memory
                        add     bootAddr,#1             ' increment hub address
                        djnz    acca,#:readByte         ' decrement count
                        andn    outa,clrMask            ' deselect flash memory
                        mov     phsb,bootSect
                        add     phsb,#12                ' offset to link
                        call    #cardSetup              ' read from flash memory
                        neg     phsb,#1
                        call    #sendRecvBits           ' read LSB
                        mov     bootSect,phsb
                        and     bootSect,#$FF
                        call    #sendRecvBits           ' read MSB
                        andn    outa,clrMask            ' deselect flash memory
                        shl     phsb,#8
                        or      bootSect,phsb           ' position and combine pieces
                        and     bootSect,wordMask
                        shl     bootSect,#12
                        cmp     bootSect,endLink wz     ' check for end of file
                if_nz   tjnz    progSize,#bootRead
nowBootSpin             rdword  acca,#dbase             ' Get dbase value
                        sub     acca,#4
                        wrlong  stackMark,acca          ' Place stack marker at dbase
                        sub     acca,#4
                        wrlong  stackMark,acca
                        rdword  accb,#vbase             ' Get vbase value
                        sub     acca,accb
                        shr     acca,#2 wz              ' Compute number of longs between
:zeroIt         if_nz   wrlong  zeroLong,accb           '  vbase and below stack marker
                if_nz   add     accb,#4
                if_nz   djnz    acca,#:zeroIt           ' Zero that space (if any)
                        tjnz    changeClk,#:checkClock  ' Change system clock?
                        wrlong  oldClkFrq,#clkfreqVal   ' No, override stored values
                        wrbyte  oldClkMod,#clksetVal
                        jmp     #:justStartUp
:checkClock             rdlong  acca,#clkfreqVal        ' Yes, get CLKFREQ value
                        cmp     acca,oldClkFrq wz       ' Is the clock frequency the same?
                        rdbyte  acca,#clksetVal         ' Get CLKSET value
                if_ne   jmp     #:changeClock           ' Is the clock mode the same also?
                        cmp     acca,oldClkMod wz       ' If both same, just go start COG
                 if_e   jmp     #:justStartUp
:changeClock            and     acca,#$F8               ' Force use of RCFAST clock while
                        clkset  acca                    '  letting requested clock start
                        mov     acca,time_xtal
:startupDelay           djnz    acca,#:startupDelay     ' Allow 20ms@20MHz to settle
                        rdbyte  acca,#clksetVal         ' Then switch to selected clock
                        clkset  acca
:justStartUp            coginit startSpin               ' Start Spin in selected cog
                        jmp     #finished               ' All done (if COGINIT fails)
                                                
cardSetup               mov     selCnt,#FLASH
                        andn    outa,clrMask            ' make sure SPI deselected
                        or      outa,clrMask
:cntCS                  or      outa,cntMask            ' pulse csCnt acca times
                        andn    outa,cntMask
                        djnz    selCnt,#:cntCS
                        movi    phsb,#%0000_0011_0      '  include read command
                        mov     selCnt,#4               ' send command & address
:sendImm                call    #sendRecvBits           ' send MSB and fill with ones
                        djnz    selCnt,#:sendImm
cardSetup_ret           ret

di                      long    9
do                      long    10
clk                     long    11
csClr                   long    25
csCnt                   long    8
selNo                   long    5
nco                     long    $1000_0000
hifreq                  long    $e0_00_00_00
freq                    long    $20_00_00_00
clockfreq               long    80_000_000
onek                    long    1000
sector                  long    512
wordMask                long    $FFFF
signMask                long    $80000000
msbMask                 long    $FF000000
diMask                  long    |< 9
doMask                  long    |< 10
clkMask                 long    |< 11
clrMask                 long    |< 25
cntMask                 long    |< 8
rdDataCmd               long    $03_000000              ' Read flash command
endLink                 long    $0FFFF000               ' Link to next sector
flashSect               long    4096 - 16               ' Amount of data in sector
stackMark               long    $FFF9FFFF               ' Stack marker
interpreter             long    ($0004 << 16) | ($F004 << 2) | %0000
time_xtal               long    20 * 20000 / 4 / 1      ' 20ms (@20MHz, 1 inst/loop)
zeroLong                long    0

selCnt                  res     1
acca                    res     1
accb                    res     1
cmdo                    res     1
cmdp                    res     1
comptr                  res     1
parptr                  res     1
parptr2                 res     1
ctr                     res     1
ctr2                    res     1
bits                    res     1
starttime               res     1
duration                res     1
bootSect                res     1
progSize                res     1
bootAddr                res     1
oldClkFrq               res     1
oldClkMod               res     1
startSpin               res     1
changeClk               res     1

{{
'  Permission is hereby granted, free of charge, to any person obtaining
'  a copy of this software and associated documentation files
'  (the "Software"), to deal in the Software without restriction,
'  including without limitation the rights to use, copy, modify, merge,
'  publish, distribute, sublicense, and/or sell copies of the Software,
'  and to permit persons to whom the Software is furnished to do so,
'  subject to the following conditions:
'
'  The above copyright notice and this permission notice shall be included
'  in all copies or substantial portions of the Software.
'
'  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
'  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
'  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
'  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
'  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
}}
